# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# References:
# DeiT: https://github.com/facebookresearch/deit
# BEiT: https://github.com/microsoft/unilm/tree/master/beit
# --------------------------------------------------------
import math
import sys
from typing import Iterable

import torch

import util.misc as misc
import util.lr_sched as lr_sched

from tqdm import tqdm

import copy
import torch.nn.functional as F
import torch.nn as nn


def train_one_epoch(model: torch.nn.Module,
                    data_loader: Iterable, optimizer: torch.optim.Optimizer,
                    device: torch.device, epoch: int, loss_scaler,
                    log_writer=None,
                    args=None):
    model.train(True)
    metric_logger = misc.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', misc.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)
    print_freq = 20

    accum_iter = args.accum_iter

    optimizer.zero_grad()

    if log_writer is not None:
        print('log_dir: {}'.format(log_writer.log_dir))

    if misc.is_main_process():
        data_loader = tqdm(data_loader, file=sys.stdout)

    for data_iter_step, (samples, _) in enumerate(metric_logger.log_every(data_loader, print_freq, header)):

        # we use a per iteration (instead of per epoch) lr scheduler
        
        """
        if data_iter_step % accum_iter == 0:
            lr_sched.adjust_learning_rate(optimizer, data_iter_step / len(data_loader) + epoch, args)
        """
        
        samples = samples.to(device, non_blocking=True)

        with torch.cuda.amp.autocast():
            loss, _, _, _, _, _, _ = model(samples, mask_ratio=args.mask_ratio)
        """
        input_1, input_2 = samples
        input_1, input_2 = input_1.to(device, non_blocking=True), input_2.to(device, non_blocking=True)
        loss = model(input_1, input_2, None, None)
        """

        if torch.cuda.device_count() > 1:
            loss = torch.mean(loss)
        loss_value = loss.item()

        if not math.isfinite(loss_value):
            print("Loss is {}, stopping training".format(loss_value))
            sys.exit(1)

        loss /= accum_iter
        loss_scaler(loss, optimizer, parameters=model.parameters(),
                    update_grad=(data_iter_step + 1) % accum_iter == 0)
        if (data_iter_step + 1) % accum_iter == 0:
            optimizer.zero_grad()

        torch.cuda.synchronize()

        metric_logger.update(loss=loss_value)

        lr = optimizer.param_groups[0]["lr"]
        metric_logger.update(lr=lr)

        loss_value_reduce = misc.all_reduce_mean(loss_value)
        if log_writer is not None and (data_iter_step + 1) % accum_iter == 0:
            """ We use epoch_1000x as the x-axis in tensorboard.
            This calibrates different curves when batch size changes.
            """
            epoch_1000x = int((data_iter_step / len(data_loader) + epoch) * 1000)
            log_writer.add_scalar('train_loss', loss_value_reduce, epoch_1000x)
            log_writer.add_scalar('lr', lr, epoch_1000x)

    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}

def train_CNN_one_epoch(model: torch.nn.Module,
                    data_loader: Iterable, optimizer: torch.optim.Optimizer,
                    device: torch.device, epoch: int, loss_scaler,
                    log_writer=None,
                    args=None):
    model.train(True)
    metric_logger = misc.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', misc.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)
    print_freq = 20

    accum_iter = args.accum_iter

    optimizer.zero_grad()

    if log_writer is not None:
        print('log_dir: {}'.format(log_writer.log_dir))

    if misc.is_main_process():
        data_loader = tqdm(data_loader, file=sys.stdout)
    
    loss_func = torch.nn.CrossEntropyLoss()

    for data_iter_step, (samples, target) in enumerate(metric_logger.log_every(data_loader, print_freq, header)):

        # input_1, input_2 = samples
        # input_1, input_2 = input_1.to(device, non_blocking=True), input_2.to(device, non_blocking=True)
        # loss = model(input_1, input_2, None, None)
        samples, target = samples.to(device, non_blocking=True), target.to(device, non_blocking=True)
        output = model(samples)
        loss = loss_func(output, target)


        if torch.cuda.device_count() > 1:
            loss = torch.mean(loss)
        loss_value = loss.item()

        if not math.isfinite(loss_value):
            print("Loss is {}, stopping training".format(loss_value))
            sys.exit(1)

        loss /= accum_iter
        loss_scaler(loss, optimizer, parameters=model.parameters(),
                    update_grad=(data_iter_step + 1) % accum_iter == 0)
        if (data_iter_step + 1) % accum_iter == 0:
            optimizer.zero_grad()

        torch.cuda.synchronize()

        metric_logger.update(loss=loss_value)

        lr = optimizer.param_groups[0]["lr"]
        metric_logger.update(lr=lr)

        loss_value_reduce = misc.all_reduce_mean(loss_value)
        if log_writer is not None and (data_iter_step + 1) % accum_iter == 0:
            """ We use epoch_1000x as the x-axis in tensorboard.
            This calibrates different curves when batch size changes.
            """
            epoch_1000x = int((data_iter_step / len(data_loader) + epoch) * 1000)
            log_writer.add_scalar('train_loss', loss_value_reduce, epoch_1000x)
            log_writer.add_scalar('lr', lr, epoch_1000x)

    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}

def KD_one_epoch(model_S: torch.nn.Module, model_T: torch.nn.Module,
                data_loader: Iterable, optimizer: torch.optim.Optimizer, device: torch.device, epoch: int, loss_scaler, 
                log_writer=None,
                args=None):
    model_S.train()
    model_T.train()
    torch.cuda.empty_cache()
    metric_logger = misc.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', misc.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)
    print_freq = 20

    accum_iter = args.accum_iter

    if log_writer is not None:
        print('log_dir: {}'.format(log_writer.log_dir))

    if misc.is_main_process():
        data_loader = tqdm(data_loader, file=sys.stdout)

    for data_iter_step, (samples, _) in enumerate(metric_logger.log_every(data_loader, print_freq, header)):

        # we use a per iteration (instead of per epoch) lr scheduler
        
        """
        if data_iter_step % accum_iter == 0:
            lr_sched.adjust_learning_rate(optimizer, data_iter_step / len(data_loader) + epoch, args)
        """
        
        samples = samples.to(device, non_blocking=True)

        optimizer.zero_grad()

        with torch.cuda.amp.autocast():
            loss, _, _, _ = model_T(samples, mask_ratio=args.mask_ratio)

        if torch.cuda.device_count() > 1:
            loss = torch.mean(loss)
        loss_value = loss.item()

        if not math.isfinite(loss_value):
            print("S Loss is {}, stopping training".format(loss_value))
            sys.exit(1)

        loss /= accum_iter 
        # loss.backward()
        # optimizer.step()
        loss_scaler(loss, optimizer, parameters=model_T.parameters(),
                    update_grad=(data_iter_step + 1) % accum_iter == 0)

        if (data_iter_step + 1) % accum_iter == 0:
            model_T.zero_grad()

        update_moving_average(model_T, model_S, epi=0.99)

        torch.cuda.synchronize()

        metric_logger.update(loss=loss_value)

        lr = optimizer.param_groups[0]["lr"]
        metric_logger.update(lr=lr)

        loss_value_reduce = misc.all_reduce_mean(loss_value)
        if log_writer is not None and (data_iter_step + 1) % accum_iter == 0:
            """ We use epoch_1000x as the x-axis in tensorboard.
            This calibrates different curves when batch size changes.
            """
            epoch_1000x = int((data_iter_step / len(data_loader) + epoch) * 1000)
            log_writer.add_scalar('train_loss', loss_value_reduce, epoch_1000x)
            log_writer.add_scalar('lr', lr, epoch_1000x)

    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}

def kd_mse_loss(logits_S, logits_T, temp=1):
    '''
    Calculate the mse loss between logits_S and logits_T
    :param logits_S: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
    :param logits_T: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
    :param temperature: A float or a tensor of shape (batch_size, length) or (batch_size,)
    '''
    if isinstance(temp, torch.Tensor) and temp.dim() > 0:
        temp = temp.unsqueeze(-1)
    beta_logits_T = F.softmax(logits_T / temp, dim=1)
    beta_logits_S = F.softmax(logits_S / temp, dim=1)
    loss = nn.MSELoss()(beta_logits_T, beta_logits_S) / 2
    return loss * temp * temp

def kd_ce_loss(logits_S, logits_T, temp=1):
    '''
    Calculate the cross entropy between logits_S and logits_T
    :param logits_S: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
    :param logits_T: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
    :param temperature: A float or a tensor of shape (batch_size, length) or (batch_size,)
    '''
    if isinstance(temp, torch.Tensor) and temp.dim() > 0:
        temp = temp.unsqueeze(-1)
    beta_logits_T = logits_T / temp
    beta_logits_S = logits_S / temp
    p_T = F.softmax(beta_logits_T, dim=-1)
    loss = -(p_T * F.log_softmax(beta_logits_S, dim=-1)).sum(dim=-1).mean()
    return loss

def hid_mse_loss(state_S, state_T, mask=None):
    '''
    * Calculates the mse loss between `state_S` and `state_T`, which are the hidden state of the models.
    * If the `inputs_mask` is given, masks the positions where ``input_mask==0``.
    * If the hidden sizes of student and teacher are different, 'proj' option is required in `inetermediate_matches` to match the dimensions.
    :param torch.Tensor state_S: tensor of shape  (*batch_size*, *length*, *hidden_size*)
    :param torch.Tensor state_T: tensor of shape  (*batch_size*, *length*, *hidden_size*)
    :param torch.Tensor mask:    tensor of shape  (*batch_size*, *length*)
    '''
    if mask is None:
        loss = F.mse_loss(state_S, state_T)
    else:
        mask = mask.to(state_S)
        valid_count = mask.sum() * state_S.size(-1)
        loss = (F.mse_loss(state_S, state_T, reduction='none') * mask.unsqueeze(-1)).sum() / valid_count
    return loss

def hid_KL_loss(state_S, state_T, temp=1):
    p_s = F.log_softmax(state_S/temp, dim=-1)
    p_t = F.softmax(state_T/temp, dim=-1)
    loss = F.kl_div(p_s, p_t, size_average=False) * (temp**2) / state_S.shape[0]
    return loss

# def update_moving_average(model_T, model_S):
#     assert isinstance(model_T, nn.Module), "Requires a teacher model for EMA model update"
#     assert isinstance(model_S, nn.Module), "Requires a student model for EMA model update"
#     student_block_id = 0
#     encoder_emas = []
#     decoder_emas = []
#     for module_name, param in model_T.named_parameters():
#         if "encoder_emas" in module_name:
#             encoder_emas.append(param)
#         if "decoder_emas" in module_name:
#             decoder_emas.append(param)
#     i = 0
#     for module_name, param in model_T.named_parameters():
#         if "blocks" in module_name and not "decoder" in module_name:
#             name_splits = module_name.split(".")
#             name_splits[2] = str(student_block_id)
#             new_name = ".".join(name_splits)
#             for S_module_name, S_param in model_S.named_parameters():
#                 if S_module_name == new_name:
#                     S_param.data = S_param.data * encoder_emas[i] + param.data * (torch.tensor(1.0) - encoder_emas[i])
#                     break
#             i += 1
#     i = 0
#     for module_name, param in model_T.named_parameters():
#         if "decoder_blocks" in module_name:
#             name_splits = module_name.split(".")
#             name_splits[2] = str(student_block_id)
#             new_name = ".".join(name_splits)
#             for S_module_name, S_param in model_S.named_parameters():
#                 if S_module_name == new_name:
#                     S_param.data = S_param.data * decoder_emas[i] + param.data * (torch.tensor(1.0) -  decoder_emas[i])
#                     break
#             i += 1

def update_moving_average(model_T, model_S, epi=0.99):
    assert isinstance(model_T, nn.Module), "Requires a teacher model for EMA model update"
    assert isinstance(model_S, nn.Module), "Requires a student model for EMA model update"
    student_block_id = 0
    for module_name, param in model_T.named_parameters():
        if "blocks" in module_name:
            name_splits = module_name.split(".")
            name_splits[2] = str(student_block_id)
            new_name = ".".join(name_splits)
            for S_module_name, S_param in model_S.named_parameters():
                if S_module_name == new_name:
                    S_param.data = S_param.data * epi + param.data * (1.0 - epi)
                    break
        else:
            for S_module_name, S_param in model_S.named_parameters():
                if S_module_name == module_name:
                    S_param.data = S_param.data * epi + param.data * (1.0 - epi)
                    break

@torch.no_grad()
def evaluate(data_loader, model, device, args):

    # switch to evaluation mode
    model.to(device)
    model.eval()

    loss_list = []
    if misc.is_main_process():
        data_loader = tqdm(data_loader, file=sys.stdout)

    for samples, _ in data_loader:

        samples = samples.to(device, non_blocking=True)

        # compute output
        with torch.cuda.amp.autocast():
            loss, _, _, _, _, _, _ = model(samples, mask_ratio=args.mask_ratio)

        # loss_list.append(loss)

        if torch.cuda.device_count() > 1:
            loss = torch.mean(loss)
        loss_value = loss.item()

    return loss_value